from micromlgen import platforms
from micromlgen.svm import is_svm, port_svm
from micromlgen.rvm import is_rvm, port_rvm
from micromlgen.sefr import is_sefr, port_sefr
from micromlgen.decisiontree import is_decisiontree, port_decisiontree
from micromlgen.randomforest import is_randomforest, port_randomforest
from micromlgen.logisticregression import is_logisticregression, port_logisticregression
from micromlgen.gaussiannb import is_gaussiannb, port_gaussiannb
from micromlgen.pca import is_pca, port_pca
from micromlgen.principalfft import is_principalfft, port_principalfft
from micromlgen.linear_regression import is_linear_regression, port_linear_regression
from micromlgen.xgboost import is_xgboost, port_xgboost

def port(
        clf,
        classname=None,
        classmap=None,
        platform=platforms.ARDUINO,
        precision=None,
        **kwargs):
    """Port a classifier to plain C++"""
    assert platform in platforms.ALL, 'Unknown platform %s. Use one of %s' % (platform, ', '.join(platforms.ALL))
    if is_svm(clf):
        return port_svm(**locals())
    elif is_rvm(clf):
        return port_rvm(**locals())
    elif is_sefr(clf):
        return port_sefr(**locals())
    elif is_decisiontree(clf):
        return port_decisiontree(**locals())
    elif is_randomforest(clf):
        return port_randomforest(**locals())
    elif is_logisticregression(clf):
        return port_logisticregression(**locals())
    elif is_gaussiannb(clf):
        return port_gaussiannb(**locals())
    elif is_pca(clf):
        return port_pca(**locals())
    elif is_principalfft(clf):
        return port_principalfft(**locals(), **kwargs)
    elif is_linear_regression(clf):
        return port_linear_regression(**locals(), **kwargs)
    elif is_xgboost(clf):
        return port_xgboost(**locals(), **kwargs)
    raise TypeError('clf MUST be one of %s' % ', '.join(platforms.ALLOWED_CLASSIFIERS))


from micromlgen.utils import jinja, check_type
from tempfile import NamedTemporaryFile
import json


def format_tree(tree):
    """
    Format xgboost tree like a sklearn DecisionTree
    :param tree:
    :return:
    """
    split_indices = tree['split_indices']
    split_conditions = tree['split_conditions']
    left_children = tree['left_children']
    right_children = tree['right_children']
    return {
        'left': left_children,
        'right': right_children,
        'features': split_indices,
        'thresholds': split_conditions
    }


def is_xgboost(clf):
    """Test if classifier can be ported"""
    return check_type(clf, 'XGBClassifier')


def port_xgboost(clf, **kwargs):
    """Port a XGBoost classifier"""
    with NamedTemporaryFile('w+', suffix='.json', encoding='utf-8') as tmp:
        clf.save_model(tmp.name)
        tmp.seek(0)
        decoded = json.load(tmp)
        trees = [format_tree(tree) for tree in decoded['learner']['gradient_booster']['model']['trees']]
        return jinja('xgboost/xgboost.jinja', {
            'n_classes': int(decoded['learner']['learner_model_param']['num_class']),
            'trees': trees,
        }, {
            'classname': 'XGBClassifier'
        }, **kwargs)


from micromlgen.utils import jinja, check_type


def is_sefr(clf):
    """Test if classifier can be ported"""
    return check_type(clf, 'SEFR')


def port_sefr(clf, classname=None, **kwargs):
    """Port SEFR classifier"""
    return jinja('sefr/sefr.jinja', {
        'weights': clf.weights,
        'bias': clf.bias,
        'dimension': len(clf.weights),
    }, {
        'classname': 'SEFR'
    }, **kwargs)


import micromlgen.platforms as platforms
from micromlgen.micromlgen import port
from micromlgen.utils import port_testset, port_trainset
from micromlgen.wifiindoorpositioning import port_wifi_indoor_positioning



from micromlgen.utils import jinja, check_type
from math import pi


def is_principalfft(clf):
    """Test if classifier can be ported"""
    return check_type(clf, 'PrincipalFFT') or check_type(clf, 'KFFT')


def port_principalfft(clf, optimize_sin=False, lookup_cos=None, lookup_sin=None, **kwargs):
    """Port PrincipalFFT classifier"""
    return jinja('principalfft/principalfft.jinja', {
        'fft': clf,
        'PI': pi,
        'size': len(clf.idx),
        'optmize_sin': optimize_sin,
        'lookup_cos': lookup_cos,
        'lookup_sin': lookup_sin,
    }, **kwargs)


import json
import numpy as np
from micromlgen.utils import jinja


def parse_samples(samples, parser):
    for line in samples.split('\n'):
        if '{' in line and '}' in line:
            data = json.loads(line)
            info = {k: v for k, v in data.items() if k.startswith('__')}
            networks = {k: v for k, v in data.items() if not k.startswith('__')}
            yield parser(info, networks)


def get_classmap(samples):
    """Get {location: classIdx} mapping"""

    def parser(info, networks):
        return info['__location']

    locations = list(parse_samples(samples, parser))
    return {location: i for i, location in enumerate(sorted(set(locations)))}


def get_networkmap(samples):
    """Get {network: featureIdx} mapping"""

    def parser(info, networks):
        return networks.keys()

    networks = [list(x) for x in parse_samples(samples, parser)]
    networks = [network for sub in networks for network in sub]
    return {network: i for i, network in enumerate(sorted(set(networks)))}


def get_x(samples, networkmap):
    """Get features array"""

    def parser(info, networks):
        x = [0] * len(networkmap)
        for network, rssi in networks.items():
            x[networkmap.get(network)] = rssi
        return x

    return np.asarray(list(parse_samples(samples, parser)), dtype=np.int8)


def get_y(samples, classmap):
    """Get locationIdx array"""

    def parser(info, networks):
        location = info['__location']
        assert location in classmap, 'Unknown location %s' % location
        return classmap[location]

    return np.asarray(list(parse_samples(samples, parser)))


def port_wifi_indoor_positioning(samples):
    classmap = get_classmap(samples)
    networkmap = get_networkmap(samples)
    X = get_x(samples, networkmap)
    y = get_y(samples, classmap)
    # classmap is flipped wrt the format `port` expects: flip it
    classmap = {v: k for k, v in classmap.items()}
    return X, y, classmap, jinja('wifiindoorpositioning/wifiindoorpositioning.jinja', {
        'X': X,
        'networkmap': networkmap
    })



from micromlgen.utils import jinja, check_type


def is_rvm(clf):
    """Test if classifier can be ported"""
    return check_type(clf, 'RVC')


def port_rvm(clf, **kwargs):
    """Port a RVM classifier"""
    return jinja('rvm/rvm.jinja', {
        'n_classes': len(clf.intercept_),
        'kernel': {
            'type': clf.kernel,
            'gamma': clf.gamma,
            'coef0': clf.coef0,
            'degree': clf.degree
        },
        'sizes': {
            'features': clf.relevant_vectors_[0].shape[1],
        },
        'arrays': {
            'vectors': clf.relevant_vectors_,
            'coefs': clf.coef_,
            'actives': clf.active_,
            'intercepts': clf.intercept_,
            'mean': clf._x_mean,
            'std': clf._x_std
        },
    }, {
        'classname': 'RVC'
    }, **kwargs)



from micromlgen.utils import jinja, check_type


def is_logisticregression(clf):
    """Test if classifier can be ported"""
    return check_type(clf, 'LogisticRegression')


def port_logisticregression(clf, **kwargs):
    """Port sklearn's LogisticRegressionClassifier"""
    return jinja('logisticregression/logisticregression.jinja', {
        'weights': clf.coef_,
        'intercept': clf.intercept_,
        'classes': clf.classes_,
        'n_classes': len(clf.classes_)
    }, {
        'classname': 'LogisticRegression'
    }, **kwargs)


from micromlgen.utils import jinja, check_type


def is_gaussiannb(clf):
    """Test if classifier can be ported"""
    return check_type(clf, 'GaussianNB')


def port_gaussiannb(clf, **kwargs):
    """Port sklearn's GaussianNB"""
    return jinja('gaussiannb/gaussiannb.jinja', {
        'sigma': clf.sigma_,
        'theta': clf.theta_,
        'prior': clf.class_prior_,
        'classes': clf.classes_,
        'n_classes': len(clf.classes_)
    }, {
        'classname': 'GaussianNB'
    }, **kwargs)


from micromlgen.utils import jinja, check_type


def is_svm(clf):
    """Test if classifier can be ported"""
    return check_type(clf, 'SVC', 'LinearSVC', 'OneClassSVM')


def port_svm(clf, **kwargs):
    """Port a SVC / LinearSVC classifier"""
    assert isinstance(clf.gamma, float), 'You probably didn\'t set an explicit value for gamma: 0.001 is a good default'
    support_v = clf.support_vectors_
    n_classes = len(clf.n_support_)
    return jinja('svm/svm.jinja', {
        'kernel': {
            'type': clf.kernel,
            'gamma': clf.gamma,
            'coef0': clf.coef0,
            'degree': clf.degree
        },
        'sizes': {
            'features': len(support_v[0]),
            'vectors': len(support_v),
            'classes': n_classes,
            'decisions': n_classes * (n_classes - 1) // 2,
            'supports': clf.n_support_
        },
        'arrays': {
            'supports': support_v,
            'intercepts': clf.intercept_,
            'coefs': clf.dual_coef_
        }
    }, {
        'classname': 'OneClassSVM' if check_type(clf, 'OneClassSVM') else 'SVM'
    }, **kwargs)


from distutils.core import setup
from glob import glob

package_data = [filename.replace('micromlgen/', '')
                for filename in glob('micromlgen/**/*.jinja', recursive=True)]

setup(
  name = 'micromlgen',
  packages = ['micromlgen'],
  version = '1.1.15',
  license='MIT',
  description = 'Generate C code for microcontrollers from Python\'s sklearn classifiers',
  author = 'Simone Salerno',
  author_email = 'eloquentarduino@gmail.com',
  url = 'https://github.com/eloquentarduino/micromlgen',
  download_url = 'https://github.com/eloquentarduino/micromlgen/dist/micromlgen-1.1.15.tar.gz',
  keywords = [
    'ML',
    'microcontrollers',
    'sklearn',
    'machine learning'
  ],
  install_requires=[
    'jinja2',
  ],
  package_data= {
    'micromlgen': package_data
  },
  classifiers=[
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Code Generators',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)


from micromlgen.utils import jinja, check_type


def is_randomforest(clf):
    """Test if classifier can be ported"""
    return check_type(clf, 'RandomForestClassifier')


def port_randomforest(clf, **kwargs):
    """Port sklearn's RandomForestClassifier"""
    return jinja('randomforest/randomforest.jinja', {
        'n_classes': clf.n_classes_,
        'trees': [{
            'left': clf.tree_.children_left,
            'right': clf.tree_.children_right,
            'features': clf.tree_.feature,
            'thresholds': clf.tree_.threshold,
            'classes': clf.tree_.value,
        } for clf in clf.estimators_]
    }, {
        'classname': 'RandomForest'
    }, **kwargs)


ARDUINO = 'arduino'
ATTINY = 'attiny'
ALL = [
    ARDUINO,
    ATTINY
]

ALLOWED_CLASSIFIERS = [
    'SVC',
    'LinearSVC',
    'OneClassSVC',
    'RVC',
    'SEFR',
    'DecisionTree',
    'RandomForest',
    'GaussianNB',
    'LogisticRegression',
    'PCA',
    'PrincipalFFT',
    'LinearRegression',
    'XGBClassifier'
]


from micromlgen.utils import jinja, check_type


def is_linear_regression(clf):
    """Test if classifier can be ported"""
    return check_type(clf, 'LinearRegression')


def port_linear_regression(clf, classname=None, **kwargs):
    """Port Linear Regression"""
    return jinja('linearregression/linearregression.jinja', {
        'coefs': clf.coef_,
        'intercept': clf.intercept_,
        'dimension': len(clf.coef_),
        'dtype': 'float'
    }, {
        'classname': 'LinearRegression'
    }, **kwargs)


from micromlgen.utils import jinja, check_type


def is_pca(clf):
    """Test if classifier can be ported"""
    return check_type(clf, 'PCA')


def port_pca(clf, **kwargs):
    """Port a PCA"""
    return jinja('pca/pca.jinja', {
        'arrays': {
            'components': clf.components_,
            'mean': clf.mean_
        },
    }, {
        'classname': 'PCA'
    }, **kwargs)


from micromlgen.utils import jinja, check_type


def is_decisiontree(clf):
    """Test if classifier can be ported"""
    return check_type(clf, 'DecisionTreeClassifier')


def port_decisiontree(clf, **kwargs):
    """Port sklearn's DecisionTreeClassifier"""
    return jinja('decisiontree/decisiontree.jinja', {
        'left': clf.tree_.children_left,
        'right': clf.tree_.children_right,
        'features': clf.tree_.feature,
        'thresholds': clf.tree_.threshold,
        'classes': clf.tree_.value,
        'i': 0
    }, {
        'classname': 'DecisionTree'
    }, **kwargs)